/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * main.c
 * Copyright (C) Bob van Loosen 2007 <loosen@xs4all.nl>
 * 
 * main.c is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * main.c is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with main.c.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 *		xine dvd:///media/iso 3>&2  2>&1  1>&3 | ./xlight
 */

#include <unistd.h>
#include <stdio.h>   /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <sys/time.h>
#include <sys/types.h>
#include "linux/input.h"
#include "keys.h"
static int uinp_fd = -1;
struct uinput_user_dev uinp;
struct input_event event;


int open_port(char* port);
int setup_uinput_device(char* uinput_node);
void send_key(int key, int value);

int main(int argc, char *argv[])
{
	unsigned char in;
	
	int fd; //serial port
	if (argc != 3)
	{
		printf("Usage: %s serial_port uinput_node\n",argv[0]);
		return 1;
	}
	
	fd = open_port(argv[1]); //open the serial port
	if (fd == -1)
	{
		printf("Can't open serial port: %s\n",argv[1]);
		return 2;
	}

	if (setup_uinput_device(argv[2]) < 0) //open de uinput device node
    {
    		return -1;
    }
	
	
	while(1)
	{
		in = ~0;
		read(fd, &in, 1); //read a byte from the serial port
		printf("%02x\n",in); //print the byte for debugging
		if (in == 0xE0) //escape code
		{
			read(fd, &in, 1);
			printf("%02x\n",in);
			switch(in & 127) //leave the most significant byte out since this signals key up or down
			{
				//up
				case 0x48: 
					send_key(KEY_UP, !(in >> 7)); //send the code for KEY_UP and the most significant bit of in
					break;
				//left
				case 0x4b:
					send_key(KEY_LEFT, !(in >> 7));
					break;
				//right
				case 0x4d:
					send_key(KEY_RIGHT, !(in >> 7));
					break;
				//down
				case 0x50:
					send_key(KEY_DOWN, !(in >> 7));
					break;
				//insert
				case 0x47:
					send_key(KEY_INSERT, !(in >> 7));
					break;
				//home
				case 0x49:
					send_key(KEY_HOME, !(in >> 7));
					break;
				//page up
				case 0x51:
					send_key(KEY_PAGEUP, !(in >> 7));
					break;
				//delete
				case 0x4f:
					send_key(KEY_DELETE, !(in >> 7));
					break;
				//end
				case 0x52:
					send_key(KEY_END, !(in >> 7));
					break;
				//page down
				case 0x53:
					send_key(KEY_PAGEDOWN, !(in >> 7));
					break;
				//keypad enter
				case 0x1C:
					send_key(KEY_KPENTER, !(in >> 7));
					break;
				//right alt
				case 0x38:
					send_key(KEY_RIGHTALT, !(in >> 7));
					break;
				//right control
				case 0x1D:
					send_key(KEY_RIGHTCTRL, !(in >> 7));
					break;
				//Fake left shift
				case 0x2a:
					//NOT MAPPED, this is sent together with print-screen, probably not needed.
					break;
				//Print Screen
				case 0x37:
					send_key(KEY_SYSRQ, !(in >> 7));
					break;
				//Ctrl-Break
				case 0x46:
					send_key(KEY_LINEFEED, !(in >> 7));
					break;

			}
		}
		else if(in == 0xE1) //escape code
		{
			//Pause
			read(fd, &in, 1);
			if ((in & 127) == 0x1D)
			{
				printf("%02x\n",in); //print the byte for debugging
				read(fd, &in, 1);
				if ((in & 127) == 0x45)
				{
					printf("%02x\n",in); //print the byte for debugging
					send_key(KEY_PAUSE, !(in >> 7));
				}
			}
		}
		else
		{
			switch(in & 127) //leave the most significant byte out since this signals key up or down
			{
				//up-left
				case 0x6E:
					//send_key(KEY_LEFT, !(in >> 7)); //send KEY_LEFT and KEY_UP, this only repeats KEY_UP for some reason.
					//send_key(KEY_UP, !(in >> 7));
					if (!(in >> 7))
					{
						send_key(KEY_UP, 1);
						send_key(KEY_LEFT, 1);
						send_key(KEY_UP, 0);
						send_key(KEY_LEFT, 0);
					}
					break;
				//help
				case 0x60:
					//NOT MAPPED
					break;
				//clear
				case 0x61:
					//NOT MAPPED
					break;
				//WS Ctrl
				case 0x62:
					//NOT MAPPED
					break;
				//Finish
				case 0x63:
					//NOT MAPPED
					break;
				//Chg Sc
				case 0x64:
					//NOT MAPPED
					break;
				//Erase EOF
				case 0x65:
					//NOT MAPPED
					break;
				//Copy Auto
				case 0x66:
					//NOT MAPPED
					break;
				//weird key with a small square, a big square and a double arrow in between
				case 0x67:
					//NOT MAPPED
					break;
				//alt-printscreen
				case 0x54:
					send_key(0x54, !(in >> 7));
					break;
				default:
					//PF13 to PF24
					if ((in & 127) >= 0x70 && (in & 127) <= 0x7B)
					{
						send_key((in & 127) + 71, !(in >> 7));
					}
					else
					
					{
						send_key(in & 127, !(in >> 7));
					}
			}
		}
		

	}
	
	close(uinp_fd); //close uinput node
	close(fd); //close serial port
	return (0);
}



int open_port(char* port)
{
	int fd; /* File descriptor for the port */
	struct termios options; //place to store the serial port options

	fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);
	if (fd != -1)
 	{
		fcntl(fd, F_SETFL, 0);
		tcgetattr(fd, &options); // get the current port options
		// set the baudrate to 9600
		cfsetispeed(&options, B9600);
		cfsetospeed(&options, B9600);
		//Enable the receiver and set local mode
		options.c_cflag |= (CLOCAL | CREAD);
		/*options.c_cflag &= ~PARENB;*/
		options.c_cflag |= (PARENB | PARODD);
		//options.c_cflag &= ~PARODD;

		options.c_cflag &= ~CSTOPB;
		
		 
		 
		options.c_cflag &= ~CSIZE;
		options.c_cflag |= CS8;
		options.c_oflag &= ~OPOST;
		options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

		options.c_iflag &= ~(IXON | IXOFF | IXANY | ICRNL   | INPCK  | IGNPAR| PARMRK);
		
		//Set the new options for the port
		tcsetattr(fd, TCSANOW, &options);
	}
	return (fd);
}

int setup_uinput_device(char* uinput_node)
{
         // Temporary variable
         int i=0;
         // Open the input device
         uinp_fd = open(uinput_node, O_WRONLY | O_NDELAY);
         if (uinp_fd == NULL)
         {
                printf("Unable to open %s\n", uinput_node);
                return -1;
         }
       memset(&uinp,0,sizeof(uinp)); // Intialize the uInput device to NULL
       strncpy(uinp.name, "PolyVision Touch Screen", UINPUT_MAX_NAME_SIZE);
       uinp.id.version = 4;
       uinp.id.bustype = BUS_USB;
       // Setup the uinput device
       ioctl(uinp_fd, UI_SET_EVBIT, EV_KEY);
       ioctl(uinp_fd, UI_SET_EVBIT, EV_REL);
       ioctl(uinp_fd, UI_SET_RELBIT, REL_X);
       ioctl(uinp_fd, UI_SET_RELBIT, REL_Y);
       for (i=0; i < 256; i++) {
                  ioctl(uinp_fd, UI_SET_KEYBIT, i);
       }
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_MOUSE);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_TOUCH);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_MOUSE);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_LEFT);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_MIDDLE);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_RIGHT);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_FORWARD);
       ioctl(uinp_fd, UI_SET_KEYBIT, BTN_BACK);
       /* Create input device into input sub-system */
       write(uinp_fd, &uinp, sizeof(uinp));
       if (ioctl(uinp_fd, UI_DEV_CREATE))
       {
                printf("Unable to create UINPUT device.\n");
                return -1;
       }
       return 1;
}
void send_key(int key, int value)
{
	int I = 0;
	
	// Report BUTTON CLICK - PRESS event
	memset(&event, 0, sizeof(event));
	gettimeofday(&event.time, NULL);
	//send the key
	event.type = EV_KEY;
	event.code = key;
	event.value = value;
	write(uinp_fd, &event, sizeof(event));
	//send SYN_REPORT so the keys actually work
	event.type = EV_SYN;
	event.code = SYN_REPORT;
	event.value = 0;
	write(uinp_fd, &event, sizeof(event));
	
	while (strcmp(keycodes[I].keyname,"Answer to Life, the Universe, and Everything") != 0 || keycodes[I].keycode != 42)
	{
		if (keycodes[I].keycode == key)
		{
			if (value == 0) printf("%s UP\n",keycodes[I].keyname);
			else printf("%s DOWN\n",keycodes[I].keyname);
			break;
		}
		I++;
	}
}
